import Ansi, * as _Ansi from './ansi.ts';
import Buffer from './buffer.ts';
import Document from './document.ts';
import Editor from './editor.ts';
import { highlightToColor, HighlightType } from './highlight.ts';
import Position from './position.ts';
import Row from './row.ts';

import * as Fn from './fns.ts';
import { defaultTerminalSize, SCROLL_TAB_SIZE } from './config.ts';
import { getTestRunner } from './runtime.ts';

const {
	assertStrictEquals: assertEquals,
	assertEquals: assertLooseEquals,
	assertExists,
	assertInstanceOf,
	assertNotEquals,
	assertNull,
	assertFalse,
	assertTrue,
	testSuite,
} = await getTestRunner();

const THIS_FILE = './src/common/all_test.ts';

// ----------------------------------------------------------------------------
// Helper Function Tests
// ----------------------------------------------------------------------------

const fnTest = () => {
	const {
		some,
		none,
		arrayInsert,
		noop,
		posSub,
		minSub,
		maxAdd,
		ord,
		strChars,
		ctrlKey,
		isControl,
		isAscii,
		isAsciiDigit,
		strlen,
		truncate,
	} = Fn;

	return {
		'some()': () => {
			assertFalse(some(null));
			assertFalse(some(void 0));
			assertFalse(some(undefined));
			assertTrue(some(0));
			assertTrue(some(false));
		},
		'none()': () => {
			assertTrue(none(null));
			assertTrue(none(void 0));
			assertTrue(none(undefined));
			assertFalse(none(0));
			assertFalse(none(false));
		},
		'arrayInsert() strings': () => {
			const a = ['😺', '😸', '😹'];
			const b = arrayInsert(a, 1, 'x');
			const c = ['😺', 'x', '😸', '😹'];
			assertLooseEquals(b, c);

			const d = arrayInsert(c, 17, 'y');
			const e = ['😺', 'x', '😸', '😹', 'y'];
			assertLooseEquals(d, e);

			assertLooseEquals(arrayInsert([], 0, 'foo'), ['foo']);
		},
		'arrayInsert() numbers': () => {
			const a = [1, 3, 5];
			const b = [1, 3, 4, 5];
			assertLooseEquals(arrayInsert(a, 2, 4), b);

			const c = [1, 2, 3, 4, 5];
			assertLooseEquals(arrayInsert(b, 1, 2), c);
		},
		'noop fn': () => {
			assertExists(noop);
			assertEquals(noop(), undefined);
		},
		'posSub()': () => {
			assertEquals(posSub(14, 15), 0);
			assertEquals(posSub(15, 1), 14);
		},
		'minSub()': () => {
			assertEquals(minSub(13, 25, -1), -1);
			assertEquals(minSub(25, 13, 0), 12);
		},
		'maxAdd()': () => {
			assertEquals(maxAdd(99, 99, 75), 75);
			assertEquals(maxAdd(25, 74, 101), 99);
		},
		'ord()': () => {
			// Invalid output
			assertEquals(ord(''), 256);

			// Valid output
			assertEquals(ord('a'), 97);
		},
		'strChars() properly splits strings into unicode characters': () => {
			assertLooseEquals(strChars('😺😸😹'), ['😺', '😸', '😹']);
		},
		'ctrlKey()': () => {
			const ctrl_a = ctrlKey('a');
			assertTrue(isControl(ctrl_a));
			assertEquals(ctrl_a, String.fromCodePoint(0x01));

			const invalid = ctrlKey('😺');
			assertFalse(isControl(invalid));
			assertEquals(invalid, '😺');
		},
		'isAscii()': () => {
			assertTrue(isAscii('asjyverkjhsdf1928374'));
			assertFalse(isAscii('😺acalskjsdf'));
			assertFalse(isAscii('ab😺ac'));
		},
		'isAsciiDigit()': () => {
			assertTrue(isAsciiDigit('1234567890'));
			assertFalse(isAsciiDigit('A1'));
			assertFalse(isAsciiDigit('/'));
			assertFalse(isAsciiDigit(':'));
		},
		'isControl()': () => {
			assertFalse(isControl('abc'));
			assertTrue(isControl(String.fromCodePoint(0x01)));
			assertFalse(isControl('😺'));
		},
		'strlen()': () => {
			// Ascii length
			assertEquals(strlen('abc'), 'abc'.length);

			// Get number of visible unicode characters
			assertEquals(strlen('😺😸😹'), 3);
			assertNotEquals('😺😸😹'.length, strlen('😺😸😹'));

			// Skin tone modifier + base character
			assertEquals(strlen('🤰🏼'), 2);
			assertNotEquals('🤰🏼'.length, strlen('🤰🏼'));

			// This has 4 sub-characters, and 3 zero-width-joiners
			assertEquals(strlen('👨‍👩‍👧‍👦'), 7);
			assertNotEquals('👨‍👩‍👧‍👦'.length, strlen('👨‍👩‍👧‍👦'));
		},
		'truncate()': () => {
			assertEquals(truncate('😺😸😹', 1), '😺');
			assertEquals(truncate('😺😸😹', 5), '😺😸😹');
			assertEquals(truncate('👨‍👩‍👧‍👦', 5), '👨‍👩‍👧');
		},
	};
};

const readKeyTest = () => {
	const { KeyCommand } = _Ansi;
	const { readKey, ctrlKey } = Fn;

	const encoder = new TextEncoder();

	const testKeyMap = (codes: string[], expected: string) => {
		codes.forEach((code) => {
			assertEquals(readKey(encoder.encode(code)), expected);
		});
	};

	return {
		'empty input': () => {
			assertEquals(readKey(new Uint8Array(0)), '');
		},
		'passthrough': () => {
			// Ignore unhandled escape sequences
			assertEquals(readKey(encoder.encode('\x1b[]')), '\x1b[]');

			// Pass explicitly mapped values right through
			assertEquals(
				readKey(encoder.encode(KeyCommand.ArrowUp)),
				KeyCommand.ArrowUp,
			);
			assertEquals(
				readKey(encoder.encode(KeyCommand.Home)),
				KeyCommand.Home,
			);
			assertEquals(
				readKey(encoder.encode(KeyCommand.Delete)),
				KeyCommand.Delete,
			);

			// And pass through whatever else
			assertEquals(readKey(encoder.encode('foobaz')), 'foobaz');
		},

		'Esc': () => testKeyMap(['\x1b', ctrlKey('l')], KeyCommand.Escape),
		'Backspace': () =>
			testKeyMap(
				[ctrlKey('h'), '\x7f'],
				KeyCommand.Backspace,
			),
		'Home': () =>
			testKeyMap(['\x1b[1~', '\x1b[7~', '\x1b[H', '\x1bOH'], KeyCommand.Home),
		'End': () =>
			testKeyMap(['\x1b[4~', '\x1b[8~', '\x1b[F', '\x1bOF'], KeyCommand.End),
		'Enter': () => testKeyMap(['\n', '\r', '\v'], KeyCommand.Enter),
	};
};

const highlightToColorTest = {
	'highlightToColor()': () => {
		assertTrue(highlightToColor(HighlightType.Number).length > 0);
		assertTrue(highlightToColor(HighlightType.Match).length > 0);
		assertTrue(highlightToColor(HighlightType.None).length > 0);
	},
};

// ----------------------------------------------------------------------------
// Tests by module
// ----------------------------------------------------------------------------

const ANSITest = () => {
	const { AnsiColor, Ground } = _Ansi;

	return {
		'color()': () => {
			assertEquals(Ansi.color(AnsiColor.FgBlue), '\x1b[34m');
		},
		'color256()': () => {
			assertEquals(Ansi.color256(128, Ground.Back), '\x1b[48;5;128m');
			assertEquals(Ansi.color256(128, Ground.Fore), '\x1b[38;5;128m');
		},
		'rgb()': () => {
			assertEquals(Ansi.rgb(32, 64, 128, Ground.Back), '\x1b[48;2;32;64;128m');
			assertEquals(Ansi.rgb(32, 64, 128, Ground.Fore), '\x1b[38;2;32;64;128m');
		},
		'moveCursor()': () => {
			assertEquals(Ansi.moveCursor(1, 2), '\x1b[2;3H');
		},
		'moveCursorForward()': () => {
			assertEquals(Ansi.moveCursorForward(2), '\x1b[2C');
		},
		'moveCursorDown()': () => {
			assertEquals(Ansi.moveCursorDown(7), '\x1b[7B');
		},
	};
};

// ----------------------------------------------------------------------------

const BufferTest = {
	'new Buffer': () => {
		const b = new Buffer();
		assertInstanceOf(b, Buffer);
		assertEquals(b.strlen(), 0);
	},
	'.appendLine': () => {
		const b = new Buffer();

		// Carriage return and line feed
		b.appendLine();
		assertEquals(b.strlen(), 2);

		b.clear();
		assertEquals(b.strlen(), 0);

		b.appendLine('foo');
		assertEquals(b.strlen(), 5);
	},
	'.append': () => {
		const b = new Buffer();

		b.append('foobar');
		assertEquals(b.strlen(), 6);
		b.clear();

		b.append('foobar', 3);
		assertEquals(b.strlen(), 3);
	},
	'.flush': async () => {
		const b = new Buffer();
		b.appendLine('foobarbaz' + Ansi.ClearLine);
		assertEquals(b.strlen(), 14);

		await b.flush();

		assertEquals(b.strlen(), 0);
	},
};

// ----------------------------------------------------------------------------

const DocumentTest = {
	'.default': () => {
		const doc = Document.default();
		assertEquals(doc.numRows, 0);
		assertTrue(doc.isEmpty());
		assertEquals(doc.row(0), null);
	},
	'.open': async () => {
		const oldDoc = Document.default();
		oldDoc.insert(Position.default(), 'foobarbaz');
		assertTrue(oldDoc.dirty);
		assertEquals(oldDoc.numRows, 1);

		const doc = await oldDoc.open(THIS_FILE);
		assertFalse(doc.dirty);
		assertFalse(doc.isEmpty());
		assertTrue(doc.numRows > 1);
	},
	'.save': async () => {
		const doc = await Document.default().open(THIS_FILE);
		doc.insertNewline(Position.default());
		assertTrue(doc.dirty);

		await doc.save('test.file');
		assertFalse(doc.dirty);
	},
	'.insertRow': () => {
		const doc = Document.default();
		doc.insertRow(undefined, 'foobar');
		assertEquals(doc.numRows, 1);
		assertFalse(doc.isEmpty());
		assertInstanceOf(doc.row(0), Row);
	},
	'.insert': () => {
		const doc = Document.default();
		assertFalse(doc.dirty);
		doc.insert(Position.at(0, 0), 'foobar');
		assertEquals(doc.numRows, 1);
		assertTrue(doc.dirty);

		doc.insert(Position.at(2, 0), 'baz');
		assertEquals(doc.numRows, 1);
		assertTrue(doc.dirty);

		doc.insert(Position.at(9, 0), 'buzz');
		assertEquals(doc.numRows, 1);
		assertTrue(doc.dirty);
		const row0 = doc.row(0);
		assertEquals(row0?.toString(), 'foobazbarbuzz');
		assertEquals(row0?.rstring(), 'foobazbarbuzz');
		assertEquals(row0?.rsize, 13);

		doc.insert(Position.at(0, 1), 'Lorem Ipsum');
		assertEquals(doc.numRows, 2);
		assertTrue(doc.dirty);
	},
	'.insertNewline': () => {
		// Invalid insert location
		const doc = Document.default();
		doc.insertNewline(Position.at(0, 3));
		assertFalse(doc.dirty);
		assertTrue(doc.isEmpty());

		// Add new empty row
		const doc2 = Document.default();
		doc2.insertNewline(Position.default());
		assertTrue(doc2.dirty);
		assertFalse(doc2.isEmpty());

		// Split an existing line
		const doc3 = Document.default();
		doc3.insert(Position.default(), 'foobar');
		doc3.insertNewline(Position.at(3, 0));
		assertEquals(doc3.numRows, 2);
		assertEquals(doc3.row(0)?.toString(), 'foo');
		assertEquals(doc3.row(1)?.toString(), 'bar');
	},
	'.delete': () => {
		const doc = Document.default();
		doc.insert(Position.default(), 'foobar');
		doc.delete(Position.at(3, 0));
		assertEquals(doc.row(0)?.toString(), 'fooar');

		// Merge previous row
		const doc2 = Document.default();
		doc2.insertNewline(Position.default());
		doc2.insert(Position.at(0, 1), 'foobar');
		doc2.delete(Position.at(0, 1));
		assertEquals(doc2.row(0)?.toString(), 'foobar');

		// Merge next row
		const doc3 = Document.default();
		doc3.insertNewline(Position.default());
		doc3.insert(Position.at(0, 1), 'foobar');
		doc3.delete(Position.at(0, 0));
		assertEquals(doc3.row(0)?.toString(), 'foobar');

		// Invalid delete location
		const doc4 = Document.default();
		doc4.insert(Position.default(), 'foobar');
		doc4.delete(Position.at(0, 1));
		assertEquals(doc4.row(0)?.toString(), 'foobar');
	},
};

// ----------------------------------------------------------------------------

const EditorTest = {
	'new Editor': () => {
		const e = new Editor(defaultTerminalSize);
		assertInstanceOf(e, Editor);
	},
	'.open': async () => {
		const e = new Editor(defaultTerminalSize);
		await e.open(THIS_FILE);
		assertInstanceOf(e, Editor);
	},
	'.processKeyPress - letters': async () => {
		const e = new Editor(defaultTerminalSize);
		const res = await e.processKeyPress('a');
		assertTrue(res);
	},
	'.processKeyPress - ctrl-q': async () => {
		const e = new Editor(defaultTerminalSize);
		const res = await e.processKeyPress(Fn.ctrlKey('q'));
		assertFalse(res);
	},
};

// ----------------------------------------------------------------------------

const PositionTest = {
	'.default': () => {
		const p = Position.default();
		assertEquals(p.x, 0);
		assertEquals(p.y, 0);
	},
	'.at': () => {
		const p = Position.at(5, 7);
		assertEquals(p.x, 5);
		assertEquals(p.y, 7);
	},
	'.from': () => {
		const p1 = Position.at(1, 2);
		const p2 = Position.from(p1);

		p1.x = 2;
		p1.y = 4;

		assertEquals(p1.x, 2);
		assertEquals(p1.y, 4);

		assertEquals(p2.x, 1);
		assertEquals(p2.y, 2);
	},
};

// ----------------------------------------------------------------------------

const RowTest = {
	'.default': () => {
		const row = Row.default();
		assertEquals(row.toString(), '');
	},
	'.from': () => {
		// From string
		const row = Row.from('xyz');
		assertEquals(row.toString(), 'xyz');

		// From existing Row
		assertEquals(Row.from(row).toString(), row.toString());

		// From 'chars'
		assertEquals(Row.from(['😺', '😸', '😹']).toString(), '😺😸😹');
	},
	'.append': () => {
		const row = Row.from('foo');
		row.append('bar');
		assertEquals(row.toString(), 'foobar');
	},
	'.delete': () => {
		const row = Row.from('foof');
		row.delete(3);
		assertEquals(row.toString(), 'foo');

		row.delete(4);
		assertEquals(row.toString(), 'foo');
	},
	'.split': () => {
		// When you split a row, it's from the cursor position
		// (Kind of like if the string were one-indexed)
		const row = Row.from('foobar');
		const row2 = Row.from('bar');
		assertEquals(row.split(3).toString(), row2.toString());
	},
	'.find': () => {
		const normalRow = Row.from('For whom the bell tolls');
		assertEquals(normalRow.find('who'), 4);
		assertNull(normalRow.find('foo'));

		const emojiRow = Row.from('😺😸😹');
		assertEquals(emojiRow.find('😹'), 2);
		assertNull(emojiRow.find('🤰🏼'));
	},
	'.byteIndexToCharIndex': () => {
		// Each 'character' is two bytes
		const row = Row.from('😺😸😹👨‍👩‍👧‍👦');
		assertEquals(row.byteIndexToCharIndex(4), 2);
		assertEquals(row.byteIndexToCharIndex(2), 1);
		assertEquals(row.byteIndexToCharIndex(0), 0);

		// Return count on nonsense index
		assertEquals(Fn.strlen(row.toString()), 10);
		assertEquals(row.byteIndexToCharIndex(72), 10);

		const row2 = Row.from('foobar');
		assertEquals(row2.byteIndexToCharIndex(2), 2);
	},
	'.charIndexToByteIndex': () => {
		// Each 'character' is two bytes
		const row = Row.from('😺😸😹👨‍👩‍👧‍👦');
		assertEquals(row.charIndexToByteIndex(2), 4);
		assertEquals(row.charIndexToByteIndex(1), 2);
		assertEquals(row.charIndexToByteIndex(0), 0);
	},
	'.cxToRx, .rxToCx': () => {
		const row = Row.from('foo\tbar\tbaz');
		row.update();
		assertNotEquals(row.chars, row.rchars);
		assertNotEquals(row.size, row.rsize);
		assertEquals(row.size, 11);
		assertEquals(row.rsize, row.size + (SCROLL_TAB_SIZE * 2) - 2);

		const cx = 11;
		const aRx = row.cxToRx(cx);
		const rx = 11;
		const aCx = row.rxToCx(aRx);
		assertEquals(aCx, cx);
		assertEquals(aRx, rx);
	},
};

// ----------------------------------------------------------------------------

const SearchTest = {};

// ----------------------------------------------------------------------------
// Test Suite Setup
// ----------------------------------------------------------------------------

testSuite({
	fns: fnTest(),
	highlightToColorTest,
	'readKey()': readKeyTest(),
	'ANSI utils': ANSITest(),
	Buffer: BufferTest,
	Document: DocumentTest,
	Editor: EditorTest,
	Position: PositionTest,
	Row: RowTest,
	Search: SearchTest,
});
